/*--------------------------------------------------------------- 
 * Copyright (c) 1999,2000,2001,2002                              
 * The Board of Trustees of the University of Illinois            
 * All Rights Reserved.                                           
 *--------------------------------------------------------------- 
 * Permission is hereby granted, free of charge, to any person    
 * obtaining a copy of this software (Iperf) and associated       
 * documentation files (the "Software"), to deal in the Software  
 * without restriction, including without limitation the          
 * rights to use, copy, modify, merge, publish, distribute,        
 * sublicense, and/or sell copies of the Software, and to permit     
 * persons to whom the Software is furnished to do
 * so, subject to the following conditions: 
 *
 *     
 * Redistributions of source code must retain the above 
 * copyright notice, this list of conditions and 
 * the following disclaimers. 
 *
 *     
 * Redistributions in binary form must reproduce the above 
 * copyright notice, this list of conditions and the following 
 * disclaimers in the documentation and/or other materials 
 * provided with the distribution. 
 * 
 *     
 * Neither the names of the University of Illinois, NCSA, 
 * nor the names of its contributors may be used to endorse 
 * or promote products derived from this Software without
 * specific prior written permission. 
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES 
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND 
 * NONINFRINGEMENT. IN NO EVENT SHALL THE CONTIBUTORS OR COPYRIGHT 
 * HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, 
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, 
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE. 
 * ________________________________________________________________
 * National Laboratory for Applied Network Research 
 * National Center for Supercomputing Applications 
 * University of Illinois at Urbana-Champaign 
 * http://www.ncsa.uiuc.edu
 * ________________________________________________________________ 
 *
 * Listener.cpp
 * by Mark Gates <mgates@nlanr.net> 
 * &  Ajay Tirumala <tirumala@ncsa.uiuc.edu> 
 * ------------------------------------------------------------------- 
 * Listener sets up a socket listening on the server host. For each 
 * connected socket that accept() returns, this creates a Server 
 * socket and spawns a thread for it. 
 * 
 * Changes to the latest version. Listener will run as a daemon 
 * Multicast Server is now Multi-threaded 
 * ------------------------------------------------------------------- 
 * headers 
 * uses 
 *   <stdlib.h> 
 *   <stdio.h> 
 *   <string.h> 
 *   <errno.h> 
 * 
 *   <sys/types.h> 
 *   <unistd.h> 
 * 
 *   <netdb.h> 
 *   <netinet/in.h> 
 *   <sys/socket.h> 
 * ------------------------------------------------------------------- */ 


#define HEADERS() 
 
#include "headers.h" 
#include "Listener.hpp" 
#include "Server.hpp" 
#include "Locale.hpp" 
#include "Settings.hpp" 
#include "PerfSocket.hpp" 
#include "util.h" 
 
 
/* ------------------------------------------------------------------- 
 * Stores local hostname and socket info. 
 * ------------------------------------------------------------------- */ 
 
using namespace std; 
#if defined(IPV6) 
vector<sockaddr_storage> clients;
#else 
vector<sockaddr> clients;
#endif
Mutex clients_mutex; 
 
Listener::Listener( short inPort, bool inUDP, const char *inLocalhost ) 
  : PerfSocket( inPort, inUDP ), 
    Thread() 
{ 
  mLocalhost = NULL; 
 
  if ( inLocalhost != NULL ) { 
    mLocalhost = new char[ strlen( inLocalhost ) + 1 ]; 
    strcpy( mLocalhost, inLocalhost ); 
  } 
 
  // open listening socket 
  Listen( mLocalhost ); 
  ReportServerSettings( inLocalhost ); 
 
} // end Listener 
 
/* ------------------------------------------------------------------- 
 * Delete memory (hostname string). 
 * ------------------------------------------------------------------- */ 
Listener::~Listener() 
{ 
  DELETE_PTR( mLocalhost  ); 
} // end ~Listener 
 
/* ------------------------------------------------------------------- 
 * Listens for connections and starts Servers to handle data. 
 * For TCP, each accepted connection spawns a Server thread. 
 * For UDP, handle all data in this thread. 
 * ------------------------------------------------------------------- */ 
void Listener::Run( void ) 
{ 
 
  struct UDP_datagram* mBuf_UDP  = (struct UDP_datagram*) mBuf; 
  if ( mUDP ) { 
    SocketAddr local = getLocalAddress(); 
    if ( local.isMulticast()) { 
      // multicast UDP uses listening socket 
      // The server will now run as a multi-threaded server 
      Server *theServer=NULL, *newServer=NULL; 
#if defined(IPV6)
      struct sockaddr_storage peer; 
#else
      struct sockaddr peer; 
#endif
      // Accept each packet, 
      // If there is no existing client, then start  
      // a new thread to service the new client 
      // The main server runs in a single thread 
      // Thread per client model is followed 
      do { 
	if (theServer==NULL) 
	      theServer = new Server(mPort, mUDP, mSock); 
        peer = get_next_UDP(); 
      
#ifndef WIN32 
        (clients_mutex).Lock(); 
        bool exist = present(peer); 
        (clients_mutex).Unlock(); 
        int32_t datagramID = ntohl( mBuf_UDP->id ); 
        if (!exist && datagramID >= 0) { 
             int rc = connect( mSock, (struct sockaddr*) &peer 
                              ,sizeof(struct sockaddr_in)); 
	     FAIL_errno( rc == SOCKET_ERROR, "connect UDP" );       
             (clients_mutex).Lock(); 
             (clients).push_back(peer); 
             (clients_mutex).Unlock(); 
	     newServer = new Server(mPort, mUDP, mSock); 
             newServer->DeleteSelfAfterRun(); 
             newServer->Start(); 
	     newServer = NULL; 
	     theServer = NULL; 
	     mSock = -1; 
             Listen(mLocalhost); 
        } 
#else 
        // WIN 32 
        theServer->Run(); 
        DELETE_PTR( theServer ); 
        // create a new socket 
        mSock = -1; 
        Listen( mLocalhost ); 
#endif 
      } while( true ); 
    } 
    else 
 { 
      // UDP uses listening socket 
      do { 
        // the server inherits control of the socket 
        Server* theServer = new Server( mPort, mUDP, mSock ); 
        theServer->Accept_UDP(); 
 
#ifndef WIN32 
        // create a new socket 
        mSock = -1; 
        Listen( mLocalhost ); 
 
        theServer->DeleteSelfAfterRun(); 
        theServer->Start(); 
        theServer = NULL; 
#else /* WIN32 */ 
        // in Windows, ALWAYS run UDP server sockets single threaded 
        // it can't handle parallel UDP streams to the same port 
        theServer->Run(); 
        DELETE_PTR( theServer ); 
 
        // create a new socket 
        mSock = -1; 
        Listen( mLocalhost ); 
#endif /* WIN32 */ 
      } while( true ); 
    } 
  } 
  else { 
    // TCP uses sockets returned from Accept 
    int connected_sock; 
    do { 
      connected_sock = Accept(); 
      if ( connected_sock >= 0 ) { 
        // startup the server thread, then forget about it 
        Server* theServer = new Server( mPort, mUDP, connected_sock ); 
 
        theServer->DeleteSelfAfterRun(); 
        theServer->Start(); 
 
        theServer = NULL; 
      } 
    } while( connected_sock != INVALID_SOCKET ); 
  } 
} // end Run 
 
 
/**-------------------------------------------------------------------- 
 * Run the server as a daemon  
 * --------------------------------------------------------------------*/ 
void Listener::runAsDaemon(const char *pname, int facility) { 
#ifndef WIN32 
	pid_t pid; 
	 
	/* Create a child process & if successful, exit from the parent process */ 
	if ((pid = fork()) == -1) { 
		printf("error in first child create\n");	 
		exit(0); 
	} else if (pid != 0) { 
		exit(0); 
	} 
	 
	/* Try becoming the session leader, once the parent exits */ 
	if (setsid() == -1) { 			/* Become the session leader */ 
		fputs("Cannot change the session group leader\n",stdout); 
	} else { 
	} 
	signal(SIGHUP,SIG_IGN); 
 
		 
	/* Now fork() and get released from the terminal */	 
	if ((pid = fork()) == -1) { 
		printf("error\n");	 
		exit(0); 
	} else if (pid != 0) { 
		exit(0); 
	} 
		 
	chdir("."); 
        printf("Running Iperf Server as a daemon\n"); 
        printf("The Iperf daemon process ID : %d\n",((int)getpid())); 
        fflush(stdout);  
	/*umask(0);*/ 
	 
	close((int)stdout); 
	close((int)stdin); 
	openlog(pname,LOG_CONS,facility); 
#else 
        printf("Use the precompiled windows version for service (daemon) option\n"); 
#endif  
 
} 
 
 
/** 
 * Function which checks whether there is an existing connection 
 * from the client which sent the most recent packet 
 */ 
#if defined(IPV6)
bool Listener::present(struct sockaddr_storage peer) {
#else
bool Listener::present(struct sockaddr peer) {
#endif  
     for (int i=0; i < (int)clients.size(); i++) { 
       struct sockaddr *sap = (struct sockaddr *)&peer, *sac = (struct sockaddr *) &clients[i];

#if defined(IPV6)       
       sockaddr_storage t1 = clients[i];
#else
       sockaddr t1 = clients[i];
#endif

#if defined(IPV6)      

       if (sap->sa_family == AF_INET6 && sac->sa_family == AF_INET6) {
	 // compare IPv6 addresses
	 sockaddr_in6* tp2 = (sockaddr_in6*) &t1;
	 sockaddr_in6* peerp = (sockaddr_in6*) &peer;
	 if (!memcmp(tp2->sin6_addr.s6_addr, peerp->sin6_addr.s6_addr, 16) &&
	     (tp2->sin6_port == peerp->sin6_port)) {
	   return true;
	 }  
       }
       else 
#endif
       if (sap->sa_family == AF_INET && sac->sa_family == AF_INET) {
	 // compare IPv4 adresses
	 sockaddr_in* tp2 = (sockaddr_in*) &t1;
	 sockaddr_in* peerp = (sockaddr_in*) &peer;
	 if ( ((long) tp2->sin_addr.s_addr == (long) peerp->sin_addr.s_addr)
	      && ( tp2->sin_port == peerp->sin_port)) {                                                   
	   return true;
	 }
       } 
     }
     return false;
}
 
 
/** 
 * Function which receives a UDP packet and returns 
 * the socket address of the peer 
 */ 
#if defined(IPV6)
struct sockaddr_storage Listener::get_next_UDP( void ) 
{ 
  struct sockaddr_storage peer; 
#else
struct sockaddr Listener::get_next_UDP( void ) 
{ 
  struct sockaddr peer; 
#endif

  Socklen_t peerlen; 
  int rc; 
     
  peerlen = sizeof(peer); 
#if defined(IPV6)
  rc = recvfrom( mSock, mBuf, mBufLen, 0, 
                   (struct sockaddr*) &peer, &peerlen ); 
#else
  rc = recvfrom( mSock, mBuf, mBufLen, 0, 
                   (struct sockaddr*) &peer, &peerlen ); 
#endif
  FAIL_errno( rc == SOCKET_ERROR, "recvfrom" );       
  return peer; 
} 
